/*
	pcxload.c

	% Load a Z-soft pcx format image file into a pixel map structure.

	5/02/89 By Ted.

	OWL 1.1
	Copyright (c) 1989 by Oakland Group, Inc.
	ALL RIGHTS RESERVED.

	Revision History:
	-----------------
*/

#include "oakhead.h"
#include "disppriv.h"
#include "frwdecl.h"
#include "pcxdecl.h"
#include "digutil.h"

OSTATIC boolean DIGPRIV pcx_loadhdr(_arg2(frw_type frw, pcxhdr_struct *pcxhdr));
OSTATIC boolean DIGPRIV pcx_loadcolmap(_arg3(frw_type frw, pcxhdr_struct *pcxhdr, ocolmap_type *cmapp));

#define FRWBUFSIZE		1024
/* -------------------------------------------------------------------------- */

void DIGPRIV dig_loadpcx(frw, pmapp, cmapp, unpackpcx)
	frw_type frw;
	pmap_type *pmapp;
	ocolmap_type *cmapp;
	pcxunpack_fptr unpackpcx;
{
	pcxhdr_struct pcxhdr;

	/* Load pcx file header */
	if (!pcx_loadhdr(frw, &pcxhdr)) {
		return;
	}
	/* Allocate Pixmap */
	*pmapp = pmap_Open(pcxhdr.xmax - pcxhdr.xmin + 1,
						pcxhdr.ymax - pcxhdr.ymin + 1);
	if (*pmapp == NULL) {
		return;
	}
	/* Allocate Buffer */
	if (!frw_openbuf(frw, FRWBUFSIZE, FRW_READMODE)) {
		pmap_Close(*pmapp);
		*pmapp = NULL;
		return;
	}
	/* Read in image */
	(*unpackpcx)(frw, &pcxhdr, *pmapp);

	/* Extract colormap */
	pcx_loadcolmap(frw, &pcxhdr, cmapp);

	frw_closebuf(frw);
	return;
}
/* -------------------------------------------------------------------------- */

OSTATIC boolean DIGPRIV pcx_loadhdr(frw, pcxhdr)
	frw_type frw;
	pcxhdr_struct *pcxhdr;
{
	/* Read in header */
	if (frw_read(frw, (VOID *)pcxhdr, sizeof(pcxhdr_struct)) < sizeof(pcxhdr_struct)) {
		return(FALSE);
	}
	pcx_hdrflip(pcxhdr);	/* Flip bytes in shorts if needed */

	/* Cursory check of header */
	if (pcxhdr->mfgr != 10 ||
		pcxhdr->encoding != 1 ||
		!(pcxhdr->version == 0 || pcxhdr->version == 2 ||
		  pcxhdr->version == 3 || pcxhdr->version == 5)) {
		return(FALSE);
	}
	return(TRUE);
}
/* -------------------------------------------------------------------------- */

OSTATIC boolean DIGPRIV pcx_loadcolmap(frw, pcxhdr, cmapp)
	frw_type frw;
	pcxhdr_struct *pcxhdr;
	ocolmap_type *cmapp;
{
	opixval ncols;

	ncols = (opixval) dig_powof2(pcxhdr->pixbits * pcxhdr->nplanes);

	if (ncols == 256 && pcxhdr->version == 5) {
		/* 256 color map at end of image */
		if (frw_readb(frw) == 12) {
			*cmapp = ocolmap_Open(0, 256);
			if (*cmapp != NULL) {
				if(frw_read(frw, (VOID *) ocolmap_entry(*cmapp, 0), 256*3) == 256*3) {
					return(TRUE);
				}
			}
		}
	}
	else {	/* 16 color color map in header */
		if (ncols <= 16 && pcxhdr->version != 3) {
			*cmapp = ocolmap_Open(0, ncols);
			if (*cmapp != NULL) {
				memmove((VOID *) ocolmap_entry(*cmapp, 0), (VOID *) pcxhdr->rgb,
						(SIZE_T)((*cmapp)->nentries * 3));
				return(TRUE);
			}
		}
	}
	return(FALSE);
}
/* -------------------------------------------------------------------------- */

boolean DIGPRIV pcx_readline(frw, obuf, linebytes, ocountp)
	frw_type frw;
	byte *obuf;
	unsigned linebytes;
	unsigned *ocountp;
{
	unsigned ibyte;
	unsigned cbyte;	/* note: hi byte used for error flag */
	byte count, xcount;

	ibyte = 0;
	count = (byte) *ocountp;
	if (count != 0) cbyte = *ocountp >> 8;

	/* For each code within a plane-line */
	for (;;) {

		if (ibyte >= linebytes) break;

		if (count == 0) {
			cbyte = frw_readb(frw);
			if (cbyte & 0xFF00) return(FALSE);

			if (((byte)cbyte & 0xC0) != 0xC0) {	/* No repeat - stash & go. */
				*obuf++ = (byte) cbyte;
				if (++ibyte >= linebytes) {
					break;	/* break out of line loop to next plane */
				}
				continue;	/* continue line loop */
			}
			else {	/* Got a loop count */
				count = (byte)cbyte & ~0xC0;

				cbyte = frw_readb(frw);
				if (cbyte & 0xFF00) return(FALSE);
			}
		}
		/* If count hangs over the end of this plane line, truncate it */
		/*  and save the remainder for the next plane */
		if (linebytes - ibyte < (unsigned) count) {
			xcount = count - (byte)(linebytes - ibyte);
			count = (byte)(linebytes - ibyte);
		}
		else  xcount = 0;

		ibyte += (unsigned) count;
		for ( ; count > 0; count--) *obuf++ = (byte) cbyte;

	}	/* end of line loop */

	/* Suck out extra byte in case of odd number of bytes in line */
	if (linebytes & 1) {
		if (xcount == 0) {
			cbyte = frw_readb(frw);
			if (cbyte & 0xFF00) return(FALSE);

			if (((byte)cbyte & 0xC0) == 0xC0) {		/* repeat... */
				xcount = (byte)cbyte & ~0xC0 - 1;

				cbyte = frw_readb(frw);
				if (cbyte & 0xFF00) return(FALSE);
			}
		}
		else xcount--;
	}
	*ocountp = xcount | (cbyte << 8);
	return(TRUE);
}
/* -------------------------------------------------------------------------- */

void DIGPRIV pcx_hdrflip(pcxhdr)
	pcxhdr_struct *pcxhdr;
/*
	Flip bytes in shorts if needed
*/
{
	if (disp_GetByteOrder() != BO_LSBYTFIRST) {
		dig_flipshort(&pcxhdr->xmin);
		dig_flipshort(&pcxhdr->xmax);
		dig_flipshort(&pcxhdr->ymin);
		dig_flipshort(&pcxhdr->ymax);
		dig_flipshort(&pcxhdr->hres);
		dig_flipshort(&pcxhdr->vres);
		dig_flipshort(&pcxhdr->linebytes);
		dig_flipshort(&pcxhdr->paltype);
	}
}
/* -------------------------------------------------------------------------- */

